#include <stdio.h>
#include <limits.h>
#include <malloc.h>

/* write little endian (for WAV) */
void write32le(int in, unsigned char * buf) {
        buf[0]=in&0xff;
        buf[1]=(in>>8)&0xff;
        buf[2]=(in>>16)&0xff;
        buf[3]=(in>>24)&0xff;
}

/* VAG decoder, eh? */

double VAG_f[5][2] = { { 0.0, 0.0 },
                   { 60.0 / 64.0, 0.0 },
		   { 115.0 / 64.0, -52.0 / 64.0 },
		   { 98.0 / 64.0, -55.0 / 64.0 } ,
		   { 122.0 / 64.00, -60.0 / 64.0 } } ;

int VAGdecodebuffer(unsigned char * input, short * out, short * hist1p, short * hist2p) {
    int predict_nr, shift_factor, flags;
    short hist1=*hist1p, hist2=*hist2p;

    int s;
    int i;
    predict_nr = input[0] >> 4;
    shift_factor = input[0] & 0xf;
    flags = input[1];
    for (i=0;i<28;i+=2) {
	s = (short)((input[i/2+2]&0xf)<<12);
	out[i]=(s >> shift_factor)+hist1*VAG_f[predict_nr][0]+hist2*VAG_f[predict_nr][1];
	hist2=hist1;
	hist1=out[i];
	s = (short)((input[i/2+2]>>4)<<12);
	out[i+1]=(s >> shift_factor)+hist1*VAG_f[predict_nr][0]+hist2*VAG_f[predict_nr][1];
	hist2=hist1;
	hist1=out[i+1];
    }

    *hist1p=hist1;
    *hist2p=hist2;

    return 0;
}

int read16(unsigned char * buf) {
	return (buf[0]<<8)|(buf[1]);
}

int read32(unsigned char * buf) {
	return (buf[0]<<24)|(buf[1]<<16)|(buf[2]<<8)|buf[3];
}

int main(int argc, char * * argv) {
	FILE * infile, * outfile;
	unsigned char buf[16];
	char outname[PATH_MAX+1]={0};
	short hist1l=0,hist2l=0,hist1r=0,hist2r;
	int srate;
	int size;
	int startoffset=0x800;
	int i,j;
	int interleave;

	if (argc!=2) {printf("usage: %s MUSIC.FAG\n",argv[0]); return 1;}

	infile=fopen(argv[1],"rb");
	if (!infile) {printf("couldn't open %s\n",argv[1]); return 1;}
	snprintf(outname,PATH_MAX,"%s.wav",argv[1]);
	outfile=fopen(outname,"wb");
	if (!outfile) {printf("couldn't open %s for output\n",outname); return 1;}

	fread(buf,1,16,infile);

	srate=22050; /*read16(buf+10);*/
	fseek(infile,0,SEEK_END);
	size=ftell(infile)-startoffset;
	interleave=0x8000;
	
	{
	char * bufl, * bufr;
        short outbufl[28];
	short outbufr[28];
        int outsize,outsizetotal;
	int offset;

        unsigned char wavhead[44] = {
                0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00,  0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20,
                0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00,  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x04, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61,  0x00, 0x00, 0x00, 0x00
        };

        outsize = size/16*28*2;
        outsizetotal = outsize+8;
        write32le(outsizetotal,wavhead+4);
        write32le(outsize,wavhead+40);
        write32le(srate,wavhead+24);
        write32le(srate*4,wavhead+28);
        if (fwrite(wavhead,1,44,outfile)!=44) return 1;

	offset=startoffset;

	bufl=malloc(sizeof(char)*interleave);
	bufr=malloc(sizeof(char)*interleave);
	fseek(infile,offset,SEEK_SET);
	for (;size>=interleave*2;size-=interleave*2) {
		//fseek(infile,offset,SEEK_SET);
		fread(bufl,1,interleave,infile);
		//fseek(infile,offset+interleave,SEEK_SET);
		fread(bufr,1,interleave,infile);
		for (i=0;i<interleave;i+=16) {
			VAGdecodebuffer(bufl+i, outbufl, &hist1l, &hist2l);
			VAGdecodebuffer(bufr+i, outbufr, &hist1r, &hist2r);
			for (j=0;j<28;j++) {
				fwrite(outbufl+j,1,2,outfile);
				fwrite(outbufr+j,1,2,outfile);
			}
		}
		offset+=interleave*2;
	}
	}
}
